"""Sync store with speech recognition engines."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, TypedDict, cast

from google_engine import GoogleSpeechRecognitionEngine
from vosk_engine import VoskEngine

from ubo_app.constants.assistant import (
    ASSISTANT_END_WORD,
    ASSISTANT_WAKE_WORD,
    INTENTS_WAKE_WORD,
)
from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.audio import AudioReportSampleEvent
from ubo_app.store.services.speech_recognition import (
    SpeechRecognitionEngineName,
    SpeechRecognitionIntent,
    SpeechRecognitionReportIntentDetectionAction,
    SpeechRecognitionReportSpeechAction,
    SpeechRecognitionReportWakeWordDetectionAction,
    SpeechRecognitionStatus,
)
from ubo_app.utils.async_ import create_task

if TYPE_CHECKING:
    from collections.abc import Sequence

    from abstraction.base_class import BaseSpeechRecognitionEngine
    from abstraction.speech_recognition_mixin import Recognition, SpeechRecognitionMixin
    from abstraction.wake_word_recognition_mixin import WakeWordRecognitionMixin

    from ubo_app.utils.types import Subscriptions


class _Engines(TypedDict):
    wake_word: WakeWordRecognitionMixin
    speech: SpeechRecognitionMixin | None


def _running_engines(engines: _Engines) -> set[BaseSpeechRecognitionEngine]:
    return cast(
        'set[BaseSpeechRecognitionEngine]',
        {engine for engine in engines.values() if engine is not None},
    )


class EnginesManager:
    """Manager for speech recognition engines."""

    def __init__(self) -> None:
        """Initialize `EnginesManager`."""
        vosk_engine = VoskEngine()
        google_engine = GoogleSpeechRecognitionEngine()
        self.engines_by_name: dict[
            SpeechRecognitionEngineName,
            SpeechRecognitionMixin,
        ] = {
            SpeechRecognitionEngineName.VOSK: vosk_engine,
            SpeechRecognitionEngineName.GOOGLE: google_engine,
        }
        self.engines: _Engines = {'wake_word': vosk_engine, 'speech': vosk_engine}
        store.autorun(lambda state: state.speech_recognition.selected_engine)(
            self._sync_selected_engine,
        )

        store.autorun(
            lambda state: (
                state.speech_recognition.is_intents_active,
                state.speech_recognition.is_assistant_active,
            ),
        )(self._sync_wake_word_engine)

        store.autorun(
            lambda state: (
                state.speech_recognition.status,
                state.speech_recognition.intents,
            ),
        )(self._sync_status)

        create_task(self._monitor_wake_word_recognitions(), name='WakeWordMonitor')
        create_task(
            self._monitor_speech_recognitions(),
            name='SpeechRecognitionMonitor',
        )

        self.subscriptions: Subscriptions = [
            store.subscribe_event(AudioReportSampleEvent, self._queue_chunk),
            self._cleanup,
        ]

    async def _queue_chunk(self, event: AudioReportSampleEvent) -> None:
        """Queue audio chunk to all running speech recognition engines."""
        for engine in _running_engines(self.engines):
            await engine.queue_audio_chunk(event.sample_speech_recognition)

    async def _sync_selected_engine(
        self,
        selected_engine: SpeechRecognitionEngineName | None,
    ) -> None:
        """Sync selected speech recognition engine."""
        if self.engines['speech'] is not None:
            await self.engines['speech'].deactivate_speech_recognition()
        self.engines['speech'] = (
            self.engines_by_name[selected_engine] if selected_engine else None
        )

    async def _sync_wake_word_engine(self, data: tuple[bool, bool]) -> None:
        """Sync wake word recognition engine based on intents and assistant status."""
        is_intents_active, is_assistant_active = data
        logger.debug(
            'Syncing recognition engine status',
            extra={
                'is_intents_active': is_intents_active,
                'is_assistant_active': is_assistant_active,
            },
        )

        if is_intents_active or is_assistant_active:
            self.engines['wake_word'].set_wake_words(
                [
                    *((INTENTS_WAKE_WORD,) if is_intents_active else ()),
                    *((ASSISTANT_WAKE_WORD,) if is_assistant_active else ()),
                ],
            )
        else:
            self.engines['wake_word'].set_wake_words(None)

    async def _sync_status(
        self,
        data: tuple[str, Sequence[SpeechRecognitionIntent]],
    ) -> None:
        """Sync speech recognition status and intents."""
        status, intents = data
        logger.debug(
            'Syncing speech recognition status',
            extra={
                'status': status,
                'intents': [intent.phrase for intent in intents],
            },
        )
        if self.engines['speech'] is None:
            logger.warning(
                'Speech recognition engine is not set, skipping status sync',
                extra={'status': status},
            )
            return
        if status is SpeechRecognitionStatus.IDLE:
            await self.engines['speech'].deactivate_speech_recognition()
        elif status is SpeechRecognitionStatus.INTENTS_WAITING:
            await self.engines['speech'].activate_speech_recognition(
                phrases=[
                    phrase.lower()
                    for intent in intents
                    for phrase in (
                        [intent.phrase]
                        if isinstance(intent.phrase, str)
                        else intent.phrase
                    )
                ],
            )
        elif status is SpeechRecognitionStatus.ASSISTANT_WAITING:
            await self.engines['speech'].activate_speech_recognition(
                end_phrase=ASSISTANT_END_WORD,
            )

    @store.with_state(
        lambda state: (
            state.speech_recognition.status,
            state.speech_recognition.intents,
        ),
    )
    def handle_speech_recognition(
        self,
        data: tuple[SpeechRecognitionStatus, Sequence[SpeechRecognitionIntent]],
        recognition: Recognition,
    ) -> None:
        """Handle speech recognitions."""
        status, intents = data
        if status is SpeechRecognitionStatus.INTENTS_WAITING:
            if intent := next(
                (
                    intent
                    for intent in intents
                    if (
                        intent.phrase.lower() == recognition.text.lower()
                        if isinstance(intent.phrase, str)
                        else recognition.text.lower()
                        in [phrase.lower() for phrase in intent.phrase]
                    )
                ),
                None,
            ):
                logger.info(
                    'Intent recognized',
                    extra={
                        'engine_name': recognition.engine_name,
                        'text': recognition.text,
                        'intent': intent,
                    },
                )
                store.dispatch(
                    SpeechRecognitionReportIntentDetectionAction(intent=intent),
                )
        elif status is SpeechRecognitionStatus.ASSISTANT_WAITING:
            logger.info(
                'Assistant command recognized',
                extra={
                    'engine_name': recognition.engine_name,
                    'text': recognition.text,
                },
            )
            store.dispatch(
                SpeechRecognitionReportSpeechAction(
                    audio=recognition.audio,
                    text=recognition.text,
                    engine_name=recognition.engine_name,
                ),
            )

    async def _monitor_wake_word_recognitions(self) -> None:
        """Monitor wake word recognitions and dispatch events."""
        while True:
            async for wake_word in self.engines['wake_word'].wake_word_recogntions():
                store.dispatch(
                    SpeechRecognitionReportWakeWordDetectionAction(wake_word=wake_word),
                )
            await asyncio.sleep(0.1)

    async def _monitor_speech_recognitions(self) -> None:
        """Monitor speech recognitions and handle them."""
        while True:
            if self.engines['speech'] is not None:
                async for recognition in self.engines['speech'].speech_recognitions():
                    self.handle_speech_recognition(recognition)
            await asyncio.sleep(0.1)

    def _cleanup(self) -> None:
        """Cleanup function to stop all engines."""
        for engine in _running_engines(self.engines):
            engine.stop()



# ruff: noqa: D100, D103
from __future__ import annotations

import asyncio
import sys
import threading
import time
import traceback
import weakref
from typing import TYPE_CHECKING, cast

from ubo_app.utils import IS_TEST_ENV
from ubo_app.utils.eeprom import read_serial_number

if TYPE_CHECKING:
    from types import TracebackType


def setup_sentry() -> None:  # pragma: no cover
    import os

    import sentry_sdk

    serial_number = read_serial_number()

    if 'SENTRY_DSN' in os.environ:
        from sentry_sdk import set_user

        sentry_sdk.init(
            traces_sample_rate=1.0,
            profiles_sample_rate=1.0,
            ignore_errors=[KeyboardInterrupt, asyncio.CancelledError],
            server_name=serial_number,
        )
        set_user({'id': serial_number})


def get_all_thread_stacks() -> dict[str, list[str]]:
    id_to_name = {th.ident: th.name for th in threading.enumerate()}
    thread_stacks = {}
    for thread_id, frame in sys._current_frames().items():  # noqa: SLF001
        thread_stacks[id_to_name.get(thread_id, f'unknown-{thread_id}')] = (
            traceback.format_stack(frame)
        )
    return thread_stacks


def get_all_thread_stacks_string() -> str:
    result = ''
    for thread_name, stack in get_all_thread_stacks().items():
        result += thread_name + '\n'
        result += '-----\n'
        result += '\n'.join(stack) + '\n'
        result += '-' * 70 + '\n'
    return result


def global_exception_handler(
    exception_type: type[BaseException],
    exception_value: BaseException,
    exception_traceback: TracebackType,
) -> None:
    _ = exception_type, exception_traceback
    from ubo_app.logger import logger

    threads_info = get_all_thread_stacks()

    logger.exception(
        'Global exception',
        exc_info=exception_value,
    )
    logger.verbose(
        'Global exception',
        extra={
            'threads': threads_info,
        },
        exc_info=exception_value,
    )


def thread_exception_handler(args: threading.ExceptHookArgs) -> None:
    from ubo_app.logger import logger

    threads_info = get_all_thread_stacks()

    _, exception_value, _, thread = args
    label = getattr(thread, 'label', None)
    service_id = getattr(thread, 'service_id', None)
    service_path = getattr(thread, 'path', None)

    logger.exception(
        'Thread exception',
        extra={
            'exception_thread': thread,
            'service': label,
            **(
                {
                    'service_id': service_id,
                }
                if service_id
                else {'service_path': service_path}
            ),
        },
        exc_info=exception_value,
    )
    logger.verbose(
        'Thread exception',
        extra={
            'exception_thread': thread,
            'threads': threads_info,
            'service': label,
            **(
                {
                    'service_id': service_id,
                }
                if service_id
                else {'service_path': service_path}
            ),
        },
        exc_info=exception_value,
    )


STACKS = weakref.WeakKeyDictionary(dict[asyncio.Task, str]())


def loop_exception_handler(
    loop: asyncio.AbstractEventLoop,
    context: dict[str, object],
) -> None:
    from ubo_app.constants import DEBUG_TASKS
    from ubo_app.logger import logger

    threads_info = get_all_thread_stacks()

    exception = context.get('exception')

    if DEBUG_TASKS:
        task = cast('asyncio.Task', context.get('future') or context.get('task'))
        parent_stack = STACKS.get(task, '<unavailable>') if task else '<unavailable>'
    else:
        parent_stack = None

    thread = threading.current_thread()
    label = getattr(thread, 'label', None)
    service_id = getattr(thread, 'service_id', None)
    service_path = getattr(thread, 'path', None)

    if exception:
        if not isinstance(exception, asyncio.CancelledError):
            logger.exception(
                'Event loop exception',
                extra={
                    'service': label,
                    **(
                        {
                            'service_id': service_id,
                        }
                        if service_id
                        else {'service_path': service_path}
                    ),
                    'loop': loop,
                    'error_message': context.get('message'),
                    'future': context.get('future'),
                    'parent_stack': parent_stack,
                },
                exc_info=cast('Exception', exception),
            )
        logger.verbose(
            'Event loop exception',
            extra={
                'service': label,
                **(
                    {
                        'service_id': service_id,
                    }
                    if service_id
                    else {'service_path': service_path}
                ),
                'loop': loop,
                'error_message': context.get('message'),
                'future': context.get('future'),
                'threads': threads_info,
                'parent_stack': parent_stack,
            },
            exc_info=cast('Exception', exception),
        )
    else:
        logger.error(
            'Event loop exception handler called without an exception in the context',
            extra={
                'service': label,
                **(
                    {
                        'service_id': service_id,
                    }
                    if service_id
                    else {'service_path': service_path}
                ),
                'loop': loop,
                'context': context,
                'parent_stack': parent_stack,
            },
        )

    if service_id is not None:
        report_service_error(
            service_id=service_id,
            exception=exception,
            context=context,
        )


def report_service_error(
    *,
    service_id: str | None = None,
    exception: object | None = None,
    context: dict[str, object] | None = None,
) -> None:
    context = context or {}

    if service_id is None:
        from ubo_app.utils.service import get_service

        service_id = get_service().service_id

    if exception is None:
        _, exception, _ = sys.exc_info()

    from ubo_app.store.settings.types import (
        ErrorReport,
        SettingsReportServiceErrorAction,
    )

    message = ''

    if isinstance(exception, Exception):
        message += '[b]exception:[/b] ' + ''.join(
            traceback.format_exception(
                type(exception),
                exception,
                exception.__traceback__,
            ),
        )

    for key in context:
        message += f'\n\n[b]{key}:[/b] {context.get(key)}'

    if IS_TEST_ENV and message == '':
        msg = 'The error could not be represented as a string, reasonably'
        raise ValueError(msg)

    try:
        from ubo_app.store.main import store

    except RuntimeError as exception_:
        if exception_.args[0] != 'Store should be created in the main thread':
            raise
    else:
        store.dispatch(
            SettingsReportServiceErrorAction(
                service_id=service_id,
                error=ErrorReport(
                    message=message,
                    timestamp=time.time(),
                ),
            ),
        )


def setup_error_handling() -> None:
    setup_sentry()
    threading.excepthook = thread_exception_handler
    sys.excepthook = global_exception_handler



"""Mixin for AI providers that require setup."""

from ubo_app.engines.abstraction.engine import EngineMixin


class AIProviderMixin(EngineMixin):
    """Base class for AI providers that require setup."""



"""Docker reducer."""

from __future__ import annotations

from dataclasses import replace
from typing import TYPE_CHECKING

from redux import (
    BaseAction,
    BaseEvent,
    CombineReducerAction,
    CombineReducerInitAction,
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
    combine_reducers,
)

from ubo_app.store.services.docker import (
    DockerAction,
    DockerImageAction,
    DockerImageEvent,
    DockerImageFetchAction,
    DockerImageFetchCompositionEvent,
    DockerImageFetchEvent,
    DockerImageRegisterAppEvent,
    DockerImageReleaseAction,
    DockerImageReleaseCompositionEvent,
    DockerImageRemoveAction,
    DockerImageRemoveCompositionEvent,
    DockerImageRemoveContainerAction,
    DockerImageRemoveContainerEvent,
    DockerImageRemoveEvent,
    DockerImageRunAction,
    DockerImageRunCompositionEvent,
    DockerImageRunContainerEvent,
    DockerImageSetDockerIdAction,
    DockerImageSetStatusAction,
    DockerImageStopAction,
    DockerImageStopCompositionEvent,
    DockerImageStopContainerEvent,
    DockerImageUpdateMetadataAction,
    DockerInstallAction,
    DockerInstallEvent,
    DockerItemStatus,
    DockerRemoveUsernameAction,
    DockerServiceState,
    DockerSetStatusAction,
    DockerStartAction,
    DockerStartEvent,
    DockerState,
    DockerStatus,
    DockerStopAction,
    DockerStopEvent,
    DockerStoreUsernameAction,
    ImageState,
)

if TYPE_CHECKING:
    from ubo_app.store.services.ip import IpUpdateInterfacesAction

Action = InitAction | DockerAction


def service_reducer(
    state: DockerServiceState | None,
    action: Action,
) -> ReducerResult[DockerServiceState, Action, BaseEvent]:
    """Docker reducer."""
    if state is None:
        if isinstance(action, InitAction):
            return DockerServiceState()
        raise InitializationActionError(action)

    match action:
        case DockerSetStatusAction():
            return replace(state, status=action.status)

        case DockerStoreUsernameAction():
            return replace(
                state,
                usernames={**state.usernames, action.registry: action.username},
            )

        case DockerRemoveUsernameAction():
            return replace(
                state,
                usernames={
                    registry: username
                    for registry, username in state.usernames.items()
                    if registry != action.registry
                },
            )

        case DockerInstallAction():
            return CompleteReducerResult(
                state=replace(state, status=DockerStatus.INSTALLING),
                events=[DockerInstallEvent()],
            )

        case DockerStartAction():
            return CompleteReducerResult(
                state=replace(state, status=DockerStatus.UNKNOWN),
                events=[DockerStartEvent()],
            )

        case DockerStopAction():
            return CompleteReducerResult(
                state=replace(state, status=DockerStatus.UNKNOWN),
                events=[DockerStopEvent()],
            )

        case _:
            return state


def image_reducer(
    state: ImageState | None,
    action: DockerImageAction | CombineReducerAction | IpUpdateInterfacesAction,
) -> ReducerResult[ImageState | None, BaseAction, BaseEvent]:
    """Image reducer."""
    if state is None:
        if (
            isinstance(action, CombineReducerInitAction)
            and action.payload
            and 'label' in action.payload
        ):
            return CompleteReducerResult(
                state=ImageState(
                    id=action.key,
                    label=action.payload['label'],
                    instructions=action.payload.get('instructions', None),
                ),
                events=[DockerImageRegisterAppEvent(image=action.key)],
            )
        raise InitializationActionError(action)

    if not isinstance(action, DockerImageAction) or action.image != state.id:
        return state

    match action:
        case DockerImageSetStatusAction():
            return replace(
                state,
                status=action.status,
                ports=action.ports if action.ports else state.ports,
                container_ip=action.ip,
            )

        case DockerImageSetDockerIdAction():
            return replace(state, docker_id=action.docker_id)

        case DockerImageUpdateMetadataAction():
            return replace(
                state,
                instructions=action.instructions
                if action.instructions is not None
                else state.instructions,
            )

        case DockerImageFetchAction():
            from docker_images import IMAGES

            if IMAGES[state.id].is_composition:
                return CompleteReducerResult(
                    state=replace(state, status=DockerItemStatus.FETCHING),
                    events=[DockerImageFetchCompositionEvent(image=state.id)],
                )
            return CompleteReducerResult(
                state=replace(state, status=DockerItemStatus.FETCHING),
                events=[DockerImageFetchEvent(image=state.id)],
            )

        case DockerImageRemoveAction():
            from docker_images import IMAGES

            if IMAGES[state.id].is_composition:
                return CompleteReducerResult(
                    state=state,
                    events=[DockerImageRemoveCompositionEvent(image=state.id)],
                )
            return CompleteReducerResult(
                state=state,
                events=[DockerImageRemoveEvent(image=state.id)],
            )

        case DockerImageRunAction():
            from docker_images import IMAGES

            if IMAGES[state.id].is_composition:
                return CompleteReducerResult(
                    state=state,
                    events=[DockerImageRunCompositionEvent(image=state.id)],
                )
            return CompleteReducerResult(
                state=state,
                events=[DockerImageRunContainerEvent(image=state.id)],
            )

        case DockerImageStopAction():
            from docker_images import IMAGES

            if IMAGES[state.id].is_composition:
                return CompleteReducerResult(
                    state=state,
                    events=[DockerImageStopCompositionEvent(image=state.id)],
                )
            return CompleteReducerResult(
                state=state,
                events=[DockerImageStopContainerEvent(image=state.id)],
            )

        case DockerImageReleaseAction():
            return CompleteReducerResult(
                state=state,
                events=[DockerImageReleaseCompositionEvent(image=state.id)],
            )

        case DockerImageRemoveContainerAction():
            return CompleteReducerResult(
                state=state,
                events=[DockerImageRemoveContainerEvent(image=state.id)],
            )

        case _:
            return state

    return state


reducer, reducer_id = combine_reducers(
    state_type=DockerState,
    action_type=DockerImageAction,
    event_type=DockerImageEvent,
    service=service_reducer,
)



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace

from redux import InitAction, InitializationActionError

from ubo_app.store.services.lightdm import (
    LightDMAction,
    LightDMClearEnabledStateAction,
    LightDMState,
    LightDMUpdateStateAction,
)


def reducer(
    state: LightDMState | None,
    action: LightDMAction | InitAction,
) -> LightDMState:
    if state is None:
        if isinstance(action, InitAction):
            return LightDMState()
        raise InitializationActionError(action)

    match action:
        case LightDMClearEnabledStateAction():
            return replace(state, is_enabled=None)

        case LightDMUpdateStateAction():
            if action.is_active is not None:
                state = replace(state, is_active=action.is_active)
            if action.is_enabled is not None:
                state = replace(state, is_enabled=action.is_enabled)
            if action.is_installed is not None:
                state = replace(state, is_installed=action.is_installed)
            if action.is_installing is not None:
                state = replace(state, is_installing=action.is_installing)
            return state

        case _:
            return state



"""Utility functions for downloading files."""

from __future__ import annotations

from typing import TYPE_CHECKING

import aiofiles
import aiohttp

from ubo_app.logger import logger

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator
    from pathlib import Path


async def download_file(
    *,
    url: str,
    path: Path,
) -> AsyncGenerator[tuple[int, int | None], None]:
    """Download a file from a URL and save it to a local path."""
    downloaded_bytes = 0
    async with (
        aiohttp.ClientSession() as session,
        session.get(url, raise_for_status=True) as response,
    ):
        total_size_header = response.headers.get('Content-Length')
        if total_size_header:
            try:
                total_size = int(total_size_header)
                if total_size <= 0:
                    logger.warning(
                        'Piper: Invalid Content-Length header',
                        extra={'header': total_size_header},
                    )
                    total_size = None
            except ValueError:
                logger.warning(
                    'Piper: Invalid Content-Length header',
                    extra={'header': total_size_header},
                )
                total_size = None
        else:
            logger.warning('Piper: No Content-Length header')
            total_size = None

        async with aiofiles.open(path, mode='wb') as f:
            async for chunk in response.content.iter_chunked(1024 * 16):
                await f.write(chunk)
                downloaded_bytes += len(chunk)
                yield (downloaded_bytes, total_size)



# ruff: noqa: D100, D103
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ubo_handle import ReducerRegistrar, register


async def setup(register_reducer: ReducerRegistrar) -> None:
    from reducer import reducer

    register_reducer(reducer)

    from setup import init_service

    init_service()


register(
    service_id='file_system',
    label='File System',
    setup=setup,
)



# ruff: noqa: D100, D101
from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING

from immutable import Immutable
from redux import BaseAction, BaseEvent

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ubo_app.store.services.ethernet import NetState


class WiFiType(StrEnum):
    WEP = 'WEP'
    WPA = 'WPA'
    WPA2 = 'WPA2'
    NOPASS = 'NOPASS'


class ConnectionState(StrEnum):
    CONNECTED = 'Connected'
    CONNECTING = 'Connecting'
    DISCONNECTED = 'Disconnected'
    UNKNOWN = 'Unknown'


class WiFiConnection(Immutable):
    ssid: str
    state: ConnectionState = ConnectionState.UNKNOWN
    signal_strength: int = 0
    password: str | None = None
    type: WiFiType | None = None
    hidden: bool = False


class WiFiAction(BaseAction): ...


class WiFiInputConnectionAction(WiFiAction): ...


class WiFiSetHasVisitedOnboardingAction(WiFiAction):
    has_visited_onboarding: bool


class WiFiUpdateAction(WiFiAction):
    connections: Sequence[WiFiConnection]
    state: NetState
    current_connection: WiFiConnection | None


class WiFiUpdateRequestAction(WiFiAction):
    reset: bool = False


class WiFiEvent(BaseEvent): ...


class WiFiInputConnectionEvent(WiFiEvent): ...


class WiFiUpdateRequestEvent(WiFiEvent): ...


class WiFiState(Immutable):
    connections: Sequence[WiFiConnection] | None
    state: NetState
    current_connection: WiFiConnection | None
    has_visited_onboarding: bool | None = None
